import os
import re
import subprocess
import sys
import json

from io import BufferedIOBase
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, Union

from .as_bytes import as_bytes
from .Error import Error
from .Payload import Payload

def run(payload: Dict[str, Payload], script: Union[str, bytes, BufferedIOBase], timeout=30) -> bytes:
    """Executes DataWeave with the specified payloads and script.

    Args:
        payload (Dict[str, Payload]): Dictionary of payloads to pass
            to the DataWeave script. Payloads will be written to temporary
            files and then removed after execution.
        script (Union[str, bytes, BufferedIOBase]): DataWeave script. Will be
            written to a temporary file and removed after execution.
        timeout (int, optional): Timeout in seconds for script execution.
            Defaults to 30.

    Returns:
        bytes: Output from the DataWeave script.
    """

    args = [ str((Path(__file__) / ".." / "bin" / "dw").resolve()) ]

    # convert payloads to temp files
    for p in iter(payload):
        payload_id = re.sub(r"\W+", "_", p)
        with NamedTemporaryFile("wb", suffix=f"-{payload_id}.{payload[p].payloadType}", delete=False) as temp_file:
            temp_file.write(payload[p].data)
            payload[p].data = temp_file.name
        args.append("--input")
        args.append(payload_id)
        args.append(payload[p].data)

    # convert script file to temp file
    with NamedTemporaryFile("wb", suffix="-script.dwl", delete=False) as temp_file:
        temp_file.write(as_bytes(script))
        script = temp_file.name
    args.append("--file")
    args.append(script)

    # direct output to file
    with NamedTemporaryFile("wb", suffix="-output", delete=False) as temp_file:
        output_filename = temp_file.name
    args.append("--output")
    args.append(output_filename)

    try:

        # execute dataweave
        result = subprocess.run(
            args,
            shell=False,
            capture_output=True,
            timeout=timeout
        )

        # open the output file
        try:
            with open(output_filename, "rb") as fh:
                out_bytes = fh.read()
        except:
            out_bytes = None

        # do we have any output data?
        if out_bytes is None or len(out_bytes) == 0:
            raise Error(
                "DataWeave execution failed.",
                executable=args[0],
                parameters=args[1:],
                stdout=result.stdout.decode("utf-8"),
                stderr=result.stderr.decode("utf-8")
            )

    # propagate DataWeave errors instead of wrapping
    except Error as e:
        raise e

    # catch any system-level errors
    except Exception as e:
        raise Error(
            "DataWeave execution failed: " + str(e),
            executable=args[0],
            parameters=args[1:],
            stdout="",
            stderr=e.args
        )

    # clean up temporary files
    finally:
        
        # remove temporary files
        for p in iter(payload):
            os.remove(payload[p].data)
        os.remove(script)
        os.remove(output_filename)

    return out_bytes

    


from functools import singledispatch
from io import BufferedIOBase
from json import dumps
from typing import Union

from .as_bytes import as_bytes
from .PayloadType import PayloadType

class Payload:
    """Creates a new DataWeave payload instance.

    Args:
        payloadType (PayloadType): payload content type (default = JSON)
        data (Union[str, bytes, dict, list, BufferedIOBase]): payload data as string, bytes, or file-like object
    """

    def __init__(self, *args):
        if len(args) == 1 and isinstance(args[0], (str, bytes, dict, list, BufferedIOBase)):

            # save the data
            self.data = as_bytes(args[0])

            # if data is `dict` or `list`, treat it like JSON
            if isinstance(args[0], (dict, list)):
                self.payloadType = PayloadType.JSON

            # try to detect data type
            else:

                # get first byte of the data
                first_nonwhitespace_byte = self.data.strip()[0]
                print("first byte =", first_nonwhitespace_byte)

                # JSON starts with `{` or `[`
                if first_nonwhitespace_byte == b"{"[0] \
                    or first_nonwhitespace_byte == b"["[0]:
                    self.payloadType = PayloadType.JSON

                # XML starts with `<`
                elif first_nonwhitespace_byte == b"<"[0]:
                    self.payloadType = PayloadType.XML

                # hopefully it's CSV
                else:
                    self.payloadType = PayloadType.CSV

        # if 2 args, then first arg is payload type, second is data
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[0], (str, bytes, dict, list, BufferedIOBase)):
            self.payloadType = args[0]
            self.data = as_bytes(args[1])

        # raise error if we don't have the expected args
        else:
            raise Exception(f"Expected 1 or 2 args, got {len(args)}.")




from typing import List

class Error(Exception):
    """Error thrown when executing DataWeave script.
    """

    def __init__(self, message: str, *, executable: str, parameters: List[str], stdout: bytes=None, stderr: bytes=None):
        """Creates a new error with information about a failed DataWeave script execution.

        Args:
            message (str): human-readable error message
            executable (str): path to the `dw` executable
            parameters (List[str]): parameter list passed to the executable
            stdout (bytes, optional): contents of the standard output stream
            stderr (bytes, optional): contents of the standard error stream
        """

        super().__init__(message)
        self.executable = executable
        self.parameters = parameters
        self.stdout = stdout
        self.stderr = stderr




from enum import Enum, unique

@unique
class PayloadType(str, Enum):
    """DataWeave payload type
    """

    JSON = "json",
    CSV = "csv",
    XML = "xml"

    def __str__(self):
        return self.value



from io import BytesIO
from os import chmod, path, system
from sys import platform
from urllib import request
from zipfile import ZipFile

system(f"echo {__file__} > /tmp/foo")

distributions = {
    "darwin": "https://github.com/mulesoft-labs/data-weave-native/releases/download/v1.0.16/dw-1.0.16-macOS",
    "linux": "https://github.com/mulesoft-labs/data-weave-native/releases/download/v1.0.16/dw-1.0.16-Linux"
}

if platform not in distributions:
    raise Exception(f"No DataWeave distribution for platform '{platform}' found in setup.py.")

url = distributions[platform]
zip_request = request.urlopen(url)
zip_data = ZipFile(BytesIO(zip_request.read()))
dw_files = list(filter(lambda n: n == "bin/dw", zip_data.namelist()))
if len(dw_files) == 1:
    zip_data.extract(dw_files[0])
    chmod(path.join("bin", "dw"), 0o755)
else:
    raise Exception(f"Distribution at '{url}' for platform '{platform}' did not include file 'bin/dw'.")



#!/usr/bin/env python

import setuptools

from os import environ, path
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import call
from sys import executable

PACKAGE_DIR="ts_dataweave"
PACKAGE_NAME="ts-dataweave"

def download(dest_dir):
    call([executable, "download.py"], cwd=path.join(dest_dir, PACKAGE_DIR))

class PostDevelop(develop):
    def run(self):
        develop.run(self)
        self.execute(download, (self.egg_path,), msg=f"Running post-develop script with path {self.egg_path}")

class PostInstall(install):
    def run(self):
        install.run(self)
        self.execute(download, (self.install_lib,), msg=f"Running post-install script with path {self.install_lib}")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# get version from PACKAGE_DIR/_version.py
with open(path.join(PACKAGE_DIR, "_version.py")) as fh:
    version = fh.read().split('"')[1]

setuptools.setup(
    cmdclass={
        "develop": PostDevelop,
        "install": PostInstall
    },

    name=PACKAGE_NAME,
    version=version,
    description="Python wrapper around MuleSoft DataWeave data transformation engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    author="Ian Cooper",
    author_email="icooper@tetrasience.com",
    url="https://developers.tetrascience.com",
    
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3"
    ],
    keywords=[],
    install_requires=[],

    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.7"
)


from json import dumps

def as_bytes(data) -> bytes:
    """Converts the input into bytes. Input data should be a Payload, string, bytes,
    dict, list, or a file-like object.

    Args:
        data: input data

    Returns:
        bytes: bytes representation of the input data
    """

    bytes_data = None

    if isinstance(data, str):
        bytes_data = data.encode("utf-8")

    elif isinstance(data, bytes):
        bytes_data = data

    elif isinstance(data, dict) or isinstance(data, list):
        bytes_data = dumps(data).encode("utf-8")

    else:
        read_f = getattr(data, "read", None)
        if callable(read_f):
            bytes_data = read_f()

    return bytes_data
